#include <stdlib.h>
#include <map>
#include <math.h>
#include <string.h>
#include <iostream>
#include "Robot.h"
#include <set>
#include <stdlib.h>
#include <stdio.h>

#define WIDTH FIELD_WIDTH
#define HEIGHT FIELD_HEIGHT
using namespace std;

struct XUnit : public Unit
{
	bool base;
	int hitpoints;
	int cooldown;
	RobotState* state;
	int damage;
	int attacker;
	bool attacker_dirty;
};

vector<XUnit*> Bases;
set<XUnit*> Robots;
XUnit** Field;
size_t* MoveRequest;
vector<RobotAction> RobotBrains;
map<int, int> DeathTime;

int BASES_COUNT;
int FIELD_WIDTH; // cells
int FIELD_HEIGHT; // cells
int MSG_SIZE; //bytes
int MEMORY_SIZE; // bytes
int SHOOT_RANGE; // cells
int ROBOT_HP; // shoots
int BASE_HP; // shoots
int RELOAD_TIME; // ticks
int VIEW_RANGE; // cells
int TRANSMIT_RANGE; // cells
int ROBOT_CONSTRUCTION_TIME; // ticks
int LAST_ITER;
int ITER;

void init_field(const char* path)
{
	if(path==NULL)
	{
		ITER=0;
		Field = new XUnit*[WIDTH*HEIGHT];
		memset(Field, 0, sizeof(XUnit*)*WIDTH*HEIGHT);
		MoveRequest = new size_t[WIDTH*HEIGHT];
		memset(MoveRequest, 0, sizeof(size_t)*WIDTH*HEIGHT);
	}
	else
	{
		FILE* fl=fopen(path, "r");
		fscanf(fl, "%d %d %d %d %d %d %d %d %d %d %d %d %d\n",
				&FIELD_WIDTH, &FIELD_HEIGHT, &MSG_SIZE, &MEMORY_SIZE, &SHOOT_RANGE,
				&ROBOT_HP, &BASE_HP, &RELOAD_TIME, &VIEW_RANGE, &TRANSMIT_RANGE,
				&ROBOT_CONSTRUCTION_TIME, &ITER, &LAST_ITER);
		fscanf(fl, "%d\n", &BASES_COUNT);
		Field = new XUnit*[WIDTH*HEIGHT];
		MoveRequest = new size_t[WIDTH*HEIGHT];
		memset(MoveRequest, 0, sizeof(size_t)*WIDTH*HEIGHT);
		for(int i=0; i<BASES_COUNT; i++)
		{
			int x, y, team, hp, cd;
			fscanf(fl, "%d %d %d %d %d\n", &x, &y, &team, &hp, &cd);
			XUnit *b = new XUnit();
			b->base=true;
			b->x = x; b->y = y;
			b->team = team;
			b->hitpoints = hp;
			b->cooldown = cd;
			b->state = NULL;
			Bases.push_back(b);
			Field[b->y*WIDTH+b->x] = b;
		}
		int robots_count;
		fscanf(fl, "%d\n", &robots_count);
		for(int i=0; i<robots_count; i++)
		{
			int x, y, team, hp, cd;
			fscanf(fl, "%d %d %d %d %d", &x, &y, &team, &hp, &cd);
			XUnit *r = new XUnit();
			r->base = false;
			r->x = x;
			r->y = y;
			r->team = team;
			r->hitpoints = hp;
			r->cooldown = cd;
			r->state = new RobotState();
			r->state->memory = operator new(MEMORY_SIZE);
			memset(r->state->memory, 0, MEMORY_SIZE);
			r->state->message = operator new(MSG_SIZE);
			memset(r->state->message, 0, MSG_SIZE);
			Robots.insert(r);
			Field[x+y*WIDTH]=r;
		}
	}
}

void save(const char* path)
{
	FILE* fl = fopen(path, "w");
	if(!fl)
	{
		cerr << "Failed to open file " << path << endl;
		abort();
	}
	fprintf(fl, "%d %d %d %d %d %d %d %d %d %d %d %d %d\n",
			FIELD_WIDTH, FIELD_HEIGHT, MSG_SIZE, MEMORY_SIZE, SHOOT_RANGE,
			ROBOT_HP, BASE_HP, RELOAD_TIME, VIEW_RANGE, TRANSMIT_RANGE,
			ROBOT_CONSTRUCTION_TIME, ITER, LAST_ITER);
	fprintf(fl, "%d\n", BASES_COUNT);
	for(int i=0; i<FIELD_HEIGHT; i++)
		for(int j=0; j<FIELD_WIDTH; j++)
			if(Field[i*WIDTH+j]!=NULL && Field[i*WIDTH+j]->base)
			{
				XUnit &b = *Field[i*WIDTH+j];
				fprintf(fl, "%d %d %d %d %d\n", j, i, b.team, b.hitpoints, b.cooldown);
			}
	fprintf(fl, "%d\n", Robots.size());
	for(int i=0; i<FIELD_HEIGHT; i++)
		for(int j=0; j<FIELD_WIDTH; j++)
			if(Field[i*WIDTH+j]!=NULL && !Field[i*WIDTH+j]->base)
			{
				XUnit &r = *Field[i*WIDTH+j];
				fprintf(fl, "%d %d %d %d %d", j, i, r.team, r.hitpoints, r.cooldown);
				for(int k=0; k<MEMORY_SIZE; k++)
					fprintf(fl, " %d", ((char*)r.state->memory)[k]);
				fprintf(fl, "\n");
			}
	fclose(fl);
}

void init_bases(int teams)
{
	if(teams>BASES_COUNT)
	{
		cerr << "Too many players (" << teams << ") for " << BASES_COUNT << " bases" << endl;
		abort();
	}

	for(int i=0; i<BASES_COUNT; i++)
	{
		XUnit *b = new XUnit();
		b->base=true;
		do
		{
			b->x = random()%WIDTH;
			b->y = random()%HEIGHT;
		} while (Field[b->y*WIDTH+b->x]!=NULL);
		b->team = (i>=teams? -1: i);
		b->hitpoints = BASE_HP;
		b->cooldown = 0;
		b->state = NULL;
		Bases.push_back(b);
		Field[b->y*WIDTH+b->x] = b;
	}
}

bool finished()
{
	if(Bases.empty())
		return true;
	for(vector<XUnit*>::iterator it = Bases.begin(); it!=Bases.end(); it++)
		DeathTime[(**it).team]=ITER;
	// exit if all bases belong to the same team
	for(vector<XUnit*>::iterator it = Bases.begin(); it!=Bases.end(); it++)
		if((**it).team != (**Bases.begin()).team)
			return false;
	return true;
}

bool shift(int x, int y, int dir, int& dx, int& dy)
{
	if(!(x>=0 && x < WIDTH && y>=0 && y<HEIGHT))
		return false;
	switch(dir%7)
	{
		case 0: dx=x; dy=y; break;
		case 1: dx=x+1; dy=y; break;
		case 2: dx=x+1-y%2; dy=y+1; break;
		case 3: dx=x-y%2; dy=y+1; break;
		case 4: dx=x-1; dy=y; break;
		case 5: dx=x-y%2; dy=y-1; break;
		case 6: dx=x+1-y%2; dy=y-1; break;
		default: return false;
	}
	return dx>=0 && dx < WIDTH && dy>=0 && dy<HEIGHT;
}

bool spawn(int x, int y, int team)
{
	// try to spawn in any direction
	for(int i=1; i<=6; i++)
	{
		int dx, dy;
		if(shift(x, y, i, dx, dy) &&
				Field[dx+dy*WIDTH]==NULL)
		{
			XUnit *r = new XUnit();
			r->base = false;
			r->x = dx;
			r->y = dy;
			r->team = team;
			r->hitpoints = ROBOT_HP;
			r->cooldown = 0;
			r->state = new RobotState();
			r->state->memory = operator new(MEMORY_SIZE);
			memset(r->state->memory, 0, MEMORY_SIZE);
			r->state->message = operator new(MSG_SIZE);
			memset(r->state->message, 0, MSG_SIZE);
			Robots.insert(r);
			Field[dx+dy*WIDTH]=r;

			return true;
		}
	}
	return false;
}

void iterate_bases()
{
	for(vector<XUnit*>::iterator it = Bases.begin(); it != Bases.end(); it++)
	{
		XUnit &b = **it;
		//cout << "Base (" << b.x << ", " << b.y << ") cd: " << b.cooldown 
			//<< ", hp: " << b.hitpoints << endl;
		if(b.team != -1)
		{
			if(b.cooldown>0)
				b.cooldown--;
			else
			{
				if(spawn(b.x, b.y, b.team) || true)
					b.cooldown = ROBOT_CONSTRUCTION_TIME;
			}
		}
	}
}

float fdist(int x1, int y1, int x2, int y2)
{
	float rx1 = x1 + 0.5f*((y1+1)%2);
	float rx2 = x2 + 0.5f*((y2+1)%2);
	return sqrt((rx2-rx1)*(rx2-rx1)+(y2-y1)*(y2-y1));
}

int get_dir(int x1, int y1, int x2, int y2)
{
	if(x1<0 || x2<0 || y1<0 || y2<0
			|| x1>=WIDTH || x2>=WIDTH || y1>=HEIGHT || y2>=HEIGHT)
		return -1;
	int dir = 0;
	float mindist=fdist(x1, y1, x2, y2);
	for(int i=1; i<=6; i++)
	{
		int dx, dy;
		if(shift(x1, y1, i, dx, dy) && fdist(dx, dy, x2, y2)<mindist)
		{
			dir = i;
			mindist = fdist(dx, dy, x2, y2);
		}
	}
	return dir;
}

int dist(int x1, int y1, int x2, int y2)
{
	if(x1<0 || x2<0 || y1<0 || y2<0
			|| x1>=WIDTH || x2>=WIDTH || y1>=HEIGHT || y2>=HEIGHT)
		return -1;

	int x=x1, y=y1;
	int d=0;
	while(x!=x2 || y!=y2)
	{
		shift(x, y, get_dir(x, y, x2, y2), x, y);
		d++;
	}
	return d;
}

void draw()
{
	cout << "iter # " << ITER << ":" << endl;
	for(int i=0; i<HEIGHT; i++)
	{
		if(i%2==0) cout << " ";
		for(int j=0; j<WIDTH; j++)
		{
			if(Field[j+i*WIDTH]==NULL)
				cout << "  ";
			else
			{
				XUnit &u = *Field[j+i*WIDTH];
				if(u.base && u.team==-1)
					cout << "@@";
				else if(u.base && u.team!=-1)
					cout << "@" << u.team;
				else
					cout << "^" << u.team;
			}
		}
		cout << endl;
	}
	usleep(50000);
}

void iterate_robots()
{
	// init intermediate data
	for(vector<XUnit*>::iterator it=Bases.begin(); it!=Bases.end(); it++)
	{
		(**it).damage=0;
		(**it).attacker = -1;
		(**it).attacker_dirty = false;
	}

	// for each robot init state (incl. transmissions)
	for(set<XUnit*>::iterator it = Robots.begin(); it != Robots.end(); it++)
	{
		XUnit& r = **it;
		RobotState& s = *r.state;

		s.x = r.x; s.y = r.y; s.team = r.team; s.hitpoints = r.hitpoints;
		s.cooldown = r.cooldown;
		s.bases.clear();
		for(vector<XUnit*>::iterator ii = Bases.begin(); ii != Bases.end(); ii++)
		{
			XUnit &b = **ii;
			if(dist(b.x, b.y, s.x, s.y)<=VIEW_RANGE)
				s.bases.push_back(b);
		}
		s.robots.clear();
		for(vector<void*>::iterator ii=s.messages.begin(); ii!=s.messages.end(); ii++)
			operator delete(*ii);
		s.messages.clear();
		r.damage = 0;
	}
	// for each robot pair, if distance allows, exchange view & messages
	for(set<XUnit*>::iterator it=Robots.begin(); it!=Robots.end(); it++)
		for(set<XUnit*>::iterator ii=Robots.begin(); ii!=Robots.end(); ii++)
			if(*ii<*it)
			{
				XUnit &r1 = **it;
				XUnit &r2 = **ii;

				int d = dist(r1.x, r1.y, r2.x, r2.y);
				if(d<VIEW_RANGE)
				{
					r1.state->robots.push_back(r2);
					r2.state->robots.push_back(r1);
				}
				if(d<TRANSMIT_RANGE)
				{
					if(r1.state->transmit)
					{
						void *copy = operator new(MSG_SIZE);
						memcpy(copy, r1.state->message, MSG_SIZE);
						r2.state->messages.push_back(copy);
					}
					if(r2.state->transmit)
					{
						void *copy = operator new(MSG_SIZE);
						memcpy(copy, r2.state->message, MSG_SIZE);
						r1.state->messages.push_back(copy);
					}
				}
			}
	// think all
	// shoot (incl. teammates, but not self)
	for(set<XUnit*>::iterator it = Robots.begin(); it != Robots.end(); it++)
	{
		XUnit &r = **it;
		RobotState &s = *r.state;
		RobotBrains[r.team](*r.state);
		if(r.cooldown==0 && s.shoot)
		{
			r.cooldown = RELOAD_TIME;
			if(!(s.shoot_x==r.x && s.shoot_y==r.y) &&
				s.shoot_x>=0 && s.shoot_x<WIDTH &&
				s.shoot_y>=0 && s.shoot_y<HEIGHT &&
				dist(s.shoot_x, s.shoot_y, r.x, r.y)<=SHOOT_RANGE &&
				Field[s.shoot_x+s.shoot_y*WIDTH]!=NULL)
			{ // hit
				XUnit &tgt = *Field[s.shoot_x+s.shoot_y*WIDTH];
				if(tgt.team!=r.team)
				{
					tgt.damage += 1;
					if(tgt.base)
					{
						if(tgt.attacker==-1 || tgt.attacker==r.team)
							tgt.attacker=r.team;
						else
							tgt.attacker_dirty = true;
					}
				}
			}
		}
		else if (r.cooldown>0)
			r.cooldown--;
	}
	// show picture
	if(DRAW && ITER%1==0) draw();
	// receive damage
	for(set<XUnit*>::iterator it = Robots.begin(); it != Robots.end(); it++)
	{
		XUnit &r = **it;
		if(r.hitpoints <= r.damage)
		{ // robot est perdu
			delete r.state;
			Robots.erase(&r);
			Field[r.x+r.y*WIDTH] = NULL;
			delete &r;
		}
		else
			r.hitpoints -= r.damage;
	}
	for(vector<XUnit*>::iterator it=Bases.begin(); it!=Bases.end(); it++)
	{
		XUnit &b = **it;
		if(b.damage>=b.hitpoints)
		{ // base captured
			if(!b.attacker_dirty)
			{
				b.hitpoints = BASE_HP;
				b.team = b.attacker;
				b.cooldown = ROBOT_CONSTRUCTION_TIME;
			}
		}
		else
			b.hitpoints -= b.damage;
	}

	// fill move intents
	memset(MoveRequest, 0, sizeof(size_t)*WIDTH*HEIGHT);
	for(set<XUnit*>::iterator it = Robots.begin(); it != Robots.end(); it++)
		if((**it).state->move_direction != 0)
		{
			XUnit &r = **it;
			int dx, dy;
			if(shift(r.x, r.y, r.state->move_direction, dx, dy)
					&& Field[dx+dy*WIDTH]==NULL)
				MoveRequest[dx+dy*WIDTH]++;
			MoveRequest[r.x + r.y*WIDTH]++; // this cell is also busy
		}
	
	// move robots
	for(set<XUnit*>::iterator it = Robots.begin(); it != Robots.end(); it++)
		if((**it).state->move_direction != 0)
		{
			XUnit &r = **it;
			int dx, dy;
			if(shift(r.x, r.y, r.state->move_direction, dx, dy)
					&& Field[dx+dy*WIDTH]==NULL
					&& MoveRequest[dx+dy*WIDTH]==1)
			{
				Field[dx+dy*WIDTH] = &r;
				Field[r.x+r.y*WIDTH] = NULL;
				r.x = dx; r.y = dy;
			}
		}
}

int free_field()
{
	cout << "I'm too lame to free memory" << endl;
}

int emulate(const vector<RobotAction>& robots, const char* in, const char* out)
{
	RobotBrains = robots;
	init_field(in);
	init_bases(robots.size());

	while(!finished() && ITER<LAST_ITER)
	{ // iterate
		ITER++;
		iterate_robots();
		iterate_bases();
	}

	int winner = -1;
	// if only one team left, it's the winner
	// if more than one teams left, the places is distributed according to bases count
	//    for each team
	// for the rest players, the places are set according to the death time
	map<int, int> bases_count;
	for(vector<XUnit*>::iterator it = Bases.begin(); it!=Bases.end(); it++)
		bases_count[(**it).team]++;

	cout << "Stopped at iter=" << ITER << endl;
	for(int i=0; i<Bases.size(); i++)
	{
		if(bases_count[i]>0)
			cout << "Team " << i << ": " << bases_count[i] << " bases" << endl;
	}
	for(map<int, int>::iterator it=DeathTime.begin(); it!=DeathTime.end(); it++)
		if(it->second != ITER && it->first!=-1)
			cout << "Team " << it->first << ": death time=" << it->second << endl;


	save(out);

	free_field();

	return winner;
}
