#pragma once
#include <vector>

using std::vector;

// model parameters
extern int BASES_COUNT;
extern int FIELD_WIDTH; // cells
extern int FIELD_HEIGHT; // cells
extern int MSG_SIZE; //bytes
extern int MEMORY_SIZE; // bytes
extern int SHOOT_RANGE; // cells
extern int ROBOT_HP; // shoots
extern int BASE_HP; // shoots
extern int RELOAD_TIME; // ticks
extern int VIEW_RANGE; // cells
extern int TRANSMIT_RANGE; // cells
extern int ROBOT_CONSTRUCTION_TIME; // ticks
extern int LAST_ITER;
extern int ITER;

#define DRAW 1


struct Unit
{
	int x, y;
	int team;
};

struct RobotState
{
	// INPUT STATE

	// Robot coordinates
	int x, y;
	int team;
	int hitpoints;
	int cooldown; // ticks unable to shoot
	// Visible robots and bases
	vector<struct Unit> robots;
	vector<struct Unit> bases;
	vector<void*> messages; // each entity points to a memory block of size MSG_SIZE

	// INPUT-OUTPUT STATE
	void* memory; // pointer to the data of local robot memory (block size is MEMORY_SIZE)

	// output state
	int move_direction; // 0: stay, 1: E, 2: SE, 3: SW, 4: W, 5: NW, 6: NE
	bool shoot;
	int shoot_x, shoot_y; // global coords
	bool transmit;
	void* message; // message points to a memory block of size MSG_SIZE
};

typedef void(*RobotAction)(RobotState&);

int emulate(const vector<RobotAction>&, const char* input_filename, const char* output_filename);

// helper functions
bool shift(int x, int y, int dir, int& dx, int& dy);
int get_dir(int x1, int y1, int x2, int y2);
